//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Shape/IMaskElementDisplay.cpp
//! @brief     Implements interface IMaskElementDisplay.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Shape/IMaskElementDisplay.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/View/Shape/SceneAdaptor.h"
#include <QAction>
#include <QGraphicsScene>
#include <QGraphicsSceneContextMenuEvent>
#include <QMenu>

namespace {

bool itemMaskValue(const MaskItemObject* item)
{
    if (const auto* maskItem = dynamic_cast<const MaskItem*>(item))
        return maskItem->maskValue();
    return false;
}

} // namespace

IMaskElementDisplay::IMaskElementDisplay(MaskItemObject* item)
    : m_adaptor(nullptr)
    , m_block_on_property_change(false)
{
    ASSERT(item);

    connect(this, &IMaskElementDisplay::xChanged, this, &IMaskElementDisplay::onChangedX,
            Qt::UniqueConnection);
    connect(this, &IMaskElementDisplay::yChanged, this, &IMaskElementDisplay::onChangedY,
            Qt::UniqueConnection);

    connect(item, &MaskItemObject::maskGeometryChanged, this,
            &IMaskElementDisplay::onGeometryChange, Qt::UniqueConnection);
    connect(item, &MaskItemObject::maskVisibilityChanged, this,
            &IMaskElementDisplay::onVisibilityChange, Qt::UniqueConnection);
}

QRectF IMaskElementDisplay::boundingRect() const
{
    return m_bounding_rect;
}

void IMaskElementDisplay::setViewSceneAdaptor(const SceneAdaptor* adaptor)
{
    ASSERT(adaptor);
    if (m_adaptor == adaptor)
        return;

    if (m_adaptor)
        disconnect(m_adaptor, &SceneAdaptor::update_request, this,
                   &IMaskElementDisplay::update_view);

    m_adaptor = adaptor;
    connect(m_adaptor, &SceneAdaptor::update_request, this, &IMaskElementDisplay::update_view,
            Qt::UniqueConnection);
    update_view();
}

void IMaskElementDisplay::paint(QPainter* painter, const QStyleOptionGraphicsItem*, QWidget*)
{
    const bool is_masked = itemMaskValue(parameterizedItem());
    painter->setBrush(MaskEditorHelper::getMaskBrush(is_masked));
    painter->setPen(MaskEditorHelper::getMaskPen(is_masked));
    painter->setRenderHints(QPainter::Antialiasing);
    painter->drawPath(maskedShape());
}

QPainterPath IMaskElementDisplay::maskedShape() const
{
    static const QSet<MaskEditorHelper::EViewTypes> relevantMaskTypes = {
        MaskEditorHelper::RECTANGLE, MaskEditorHelper::POLYGON, MaskEditorHelper::VERTICALLINE,
        MaskEditorHelper::HORIZONTALLINE, MaskEditorHelper::ELLIPSE};

    QPainterPath result = mapToScene(shape());
    for (const QGraphicsItem* item : scene()->items()) {
        const auto* view = dynamic_cast<const IMaskElementDisplay*>(item);
        if (!view)
            continue;
        if (!relevantMaskTypes.contains((MaskEditorHelper::EViewTypes)view->type()))
            continue;
        if (itemMaskValue(view->parameterizedItem()))
            continue;
        if (const auto* mask_item = dynamic_cast<const MaskItem*>(view->parameterizedItem()))
            if (!mask_item->isVisible())
                continue;
        if (!(view->zValue() > zValue()))
            continue;

        const QPainterPath mask_shape = view->mapToScene(view->shape());
        if (!mask_shape.intersects(result))
            continue;

        // Item lays on top and is non-masking -> subtract the path of the item
        result = result.subtracted(mask_shape);
    }

    return mapFromScene(result);
}

qreal IMaskElementDisplay::toSceneX(qreal value) const
{
    return m_adaptor ? m_adaptor->toSceneX(value) : value;
}

qreal IMaskElementDisplay::toSceneY(qreal value) const
{
    return m_adaptor ? m_adaptor->toSceneY(value) : value;
}

qreal IMaskElementDisplay::fromSceneX(qreal value) const
{
    return m_adaptor ? m_adaptor->fromSceneX(value) : value;
}

qreal IMaskElementDisplay::fromSceneY(qreal value) const
{
    return m_adaptor ? m_adaptor->fromSceneY(value) : value;
}

void IMaskElementDisplay::addView(IMaskElementDisplay* childView)
{
    if (!childItems().contains(childView))
        childView->setParentItem(this);
}

void IMaskElementDisplay::setBlockOnProperty(bool value)
{
    m_block_on_property_change = value;
}

void IMaskElementDisplay::onGeometryChange()
{
    if (m_block_on_property_change)
        return;

    m_block_on_property_change = true;
    onPropertyChange();
    m_block_on_property_change = false;
}

void IMaskElementDisplay::onVisibilityChange()
{
    if (m_block_on_property_change)
        return;

    m_block_on_property_change = true;

    if (auto* maskItem = dynamic_cast<MaskItem*>(parameterizedItem()))
        this->setVisible(maskItem->isVisible());
    else
        this->setVisible(false);

    onPropertyChange();
    update();

    m_block_on_property_change = false;
}
